@@ -21,34 +21,71 @@
 import errno
 import os
 import re
+import uuid
 
-from lxml import etree
+import os_traits
 from oslo_concurrency import processutils
 from oslo_log import log as logging
+from oslo_utils import fileutils
 
-from nova.compute import arch
-from nova.compute import vm_mode
 import nova.conf
 from nova.i18n import _
-from nova.i18n import _LI
+from nova import objects
+from nova.objects import fields as obj_fields
+import nova.privsep.fs
+import nova.privsep.idmapshift
+import nova.privsep.libvirt
+from nova.scheduler import utils as scheduler_utils
 from nova import utils
 from nova.virt import images
 from nova.virt.libvirt import config as vconfig
 from nova.virt.libvirt.volume import remotefs
-from nova.virt import volumeutils
 
 CONF = nova.conf.CONF
 LOG = logging.getLogger(__name__)
 
 RESIZE_SNAPSHOT_NAME = 'nova-resize'
 
+# Mapping used to convert libvirt cpu features to traits, for more details, see
+# https://github.com/libvirt/libvirt/blob/master/src/cpu_map/
+CPU_TRAITS_MAPPING = {
+    '3dnow': os_traits.HW_CPU_X86_3DNOW,
+    'abm': os_traits.HW_CPU_X86_ABM,
+    'aes': os_traits.HW_CPU_X86_AESNI,
+    'avx': os_traits.HW_CPU_X86_AVX,
+    'avx2': os_traits.HW_CPU_X86_AVX2,
+    'avx512bw': os_traits.HW_CPU_X86_AVX512BW,
+    'avx512cd': os_traits.HW_CPU_X86_AVX512CD,
+    'avx512dq': os_traits.HW_CPU_X86_AVX512DQ,
+    'avx512er': os_traits.HW_CPU_X86_AVX512ER,
+    'avx512f': os_traits.HW_CPU_X86_AVX512F,
+    'avx512pf': os_traits.HW_CPU_X86_AVX512PF,
+    'avx512vl': os_traits.HW_CPU_X86_AVX512VL,
+    'avx512vnni': os_traits.HW_CPU_X86_AVX512VNNI,
+    'bmi1': os_traits.HW_CPU_X86_BMI,
+    'bmi2': os_traits.HW_CPU_X86_BMI2,
+    'pclmuldq': os_traits.HW_CPU_X86_CLMUL,
+    'f16c': os_traits.HW_CPU_X86_F16C,
+    'fma': os_traits.HW_CPU_X86_FMA3,
+    'fma4': os_traits.HW_CPU_X86_FMA4,
+    'mmx': os_traits.HW_CPU_X86_MMX,
+    'mpx': os_traits.HW_CPU_X86_MPX,
+    'sha-ni': os_traits.HW_CPU_X86_SHA,
+    'sse': os_traits.HW_CPU_X86_SSE,
+    'sse2': os_traits.HW_CPU_X86_SSE2,
+    'sse3': os_traits.HW_CPU_X86_SSE3,
+    'sse4.1': os_traits.HW_CPU_X86_SSE41,
+    'sse4.2': os_traits.HW_CPU_X86_SSE42,
+    'sse4a': os_traits.HW_CPU_X86_SSE4A,
+    'ssse3': os_traits.HW_CPU_X86_SSSE3,
+    'svm': os_traits.HW_CPU_X86_SVM,
+    'tbm': os_traits.HW_CPU_X86_TBM,
+    'vmx': os_traits.HW_CPU_X86_VMX,
+    'xop': os_traits.HW_CPU_X86_XOP
+}
 
-def execute(*args, **kwargs):
-    return utils.execute(*args, **kwargs)
-
-
-def get_iscsi_initiator():
-    return volumeutils.get_iscsi_initiator()
+# Reverse CPU_TRAITS_MAPPING
+TRAITS_CPU_MAPPING = {v: k for k, v in CPU_TRAITS_MAPPING.items()}
 
 
 def create_image(disk_format, path, size):
@@ -63,7 +100,7 @@ def create_image(disk_format, path, size):
                  M for Mebibytes, 'G' for Gibibytes, 'T' for Tebibytes).
                  If no suffix is given, it will be interpreted as bytes.
     """
-    execute('qemu-img', 'create', '-f', disk_format, path, size)
+    processutils.execute('qemu-img', 'create', '-f', disk_format, path, size)
 
 
 def create_cow_image(backing_file, path, size=None):
@@ -95,7 +132,28 @@ def create_cow_image(backing_file, path, size=None):
         csv_opts = ",".join(cow_opts)
         cow_opts = ['-o', csv_opts]
     cmd = base_cmd + cow_opts + [path]
-    execute(*cmd)
+    processutils.execute(*cmd)
+
+
+def create_ploop_image(disk_format, path, size, fs_type):
+    """Create ploop image
+
+    :param disk_format: Disk image format (as known by ploop)
+    :param path: Desired location of the ploop image
+    :param size: Desired size of ploop image. May be given as an int or
+                 a string. If given as an int, it will be interpreted
+                 as bytes. If it's a string, it should consist of a number
+                 with an optional suffix ('K' for Kibibytes,
+                 M for Mebibytes, 'G' for Gibibytes, 'T' for Tebibytes).
+                 If no suffix is given, it will be interpreted as bytes.
+    :param fs_type: Filesystem type
+    """
+    if not fs_type:
+        fs_type = CONF.default_ephemeral_format or \
+                  nova.privsep.fs.FS_FORMAT_EXT4
+    fileutils.ensure_tree(path)
+    disk_path = os.path.join(path, 'root.hds')
+    nova.privsep.libvirt.ploop_init(size, disk_format, fs_type, disk_path)
 
 
 def pick_disk_driver_name(hypervisor_version, is_block_dev=False):
@@ -117,8 +175,7 @@ def pick_disk_driver_name(hypervisor_version, is_block_dev=False):
             # 4002000 == 4.2.0
             if hypervisor_version >= 4002000:
                 try:
-                    execute('xend', 'status',
-                            run_as_root=True, check_exit_code=True)
+                    nova.privsep.libvirt.xend_probe()
                 except OSError as exc:
                     if exc.errno == errno.ENOENT:
                         LOG.debug("xend is not found")
@@ -132,7 +189,8 @@ def pick_disk_driver_name(hypervisor_version, is_block_dev=False):
                     return 'qemu'
             # libvirt will use xend/xm toolstack
             try:
-                out, err = execute('tap-ctl', 'check', check_exit_code=False)
+                out, err = processutils.execute('tap-ctl', 'check',
+                                                check_exit_code=False)
                 if out == 'ok\n':
                     # 4000000 == 4.0.0
                     if hypervisor_version > 4000000:
@@ -140,7 +198,7 @@ def pick_disk_driver_name(hypervisor_version, is_block_dev=False):
                     else:
                         return "tap"
                 else:
-                    LOG.info(_LI("tap-ctl check: %s"), out)
+                    LOG.info("tap-ctl check: %s", out)
             except OSError as exc:
                 if exc.errno == errno.ENOENT:
                     LOG.debug("tap-ctl tool is not installed")
@@ -200,7 +258,7 @@ def copy_image(src, dest, host=None, receive=False,
         # rather recreated efficiently.  In addition, since
         # coreutils 8.11, holes can be read efficiently too.
         # we add '-r' argument because ploop disks are directories
-        execute('cp', '-r', src, dest)
+        processutils.execute('cp', '-r', src, dest)
     else:
         if receive:
             src = "%s:%s" % (utils.safe_ip_format(host), src)
@@ -213,43 +271,14 @@ def copy_image(src, dest, host=None, receive=False,
             compression=compression)
 
 
-def write_to_file(path, contents, umask=None):
+def write_to_file(path, contents):
     """Write the given contents to a file
 
     :param path: Destination file
     :param contents: Desired contents of the file
-    :param umask: Umask to set when creating this file (will be reset)
     """
-    if umask:
-        saved_umask = os.umask(umask)
-
-    try:
-        with open(path, 'w') as f:
-            f.write(contents)
-    finally:
-        if umask:
-            os.umask(saved_umask)
-
-
-def chown(path, owner):
-    """Change ownership of file or directory
-
-    :param path: File or directory whose ownership to change
-    :param owner: Desired new owner (given as uid or username)
-    """
-    execute('chown', owner, path, run_as_root=True)
-
-
-def update_mtime(path):
-    """Touch a file without being the owner.
-
-    :param path: File bump the mtime on
-    """
-    execute('touch', '-c', path, run_as_root=True)
-
-
-def _id_map_to_config(id_map):
-    return "%s:%s:%s" % (id_map.start, id_map.target, id_map.count)
+    with open(path, 'w') as f:
+        f.write(contents)
 
 
 def chown_for_id_maps(path, id_maps):
@@ -259,14 +288,11 @@ def chown_for_id_maps(path, id_maps):
     :param path: File or directory whose ownership to change
     :param id_maps: List of type LibvirtConfigGuestIDMap
     """
-    uid_maps_str = ','.join([_id_map_to_config(id_map) for id_map in id_maps if
-                             isinstance(id_map,
-                                        vconfig.LibvirtConfigGuestUIDMap)])
-    gid_maps_str = ','.join([_id_map_to_config(id_map) for id_map in id_maps if
-                             isinstance(id_map,
-                                        vconfig.LibvirtConfigGuestGIDMap)])
-    execute('nova-idmapshift', '-i', '-u', uid_maps_str,
-            '-g', gid_maps_str, path, run_as_root=True)
+    uid_maps = [id_map for id_map in id_maps if
+                isinstance(id_map, vconfig.LibvirtConfigGuestUIDMap)]
+    gid_maps = [id_map for id_map in id_maps if
+                isinstance(id_map, vconfig.LibvirtConfigGuestGIDMap)]
+    nova.privsep.idmapshift.shift(path, uid_maps, gid_maps)
 
 
 def extract_snapshot(disk_path, source_fmt, out_path, dest_fmt):
@@ -282,14 +308,9 @@ def extract_snapshot(disk_path, source_fmt, out_path, dest_fmt):
     if dest_fmt == 'ploop':
         dest_fmt = 'parallels'
 
-    qemu_img_cmd = ('qemu-img', 'convert', '-f', source_fmt, '-O', dest_fmt)
-
-    # Conditionally enable compression of snapshots.
-    if CONF.libvirt.snapshot_compression and dest_fmt == "qcow2":
-        qemu_img_cmd += ('-c',)
-
-    qemu_img_cmd += (disk_path, out_path)
-    execute(*qemu_img_cmd)
+    compress = CONF.libvirt.snapshot_compression and dest_fmt == "qcow2"
+    images.convert_image(disk_path, out_path, source_fmt, dest_fmt,
+                         compress=compress)
 
 
 def load_file(path):
@@ -313,57 +334,33 @@ def file_open(*args, **kwargs):
     return open(*args, **kwargs)
 
 
-def file_delete(path):
-    """Delete (unlink) file
-
-    Note: The reason this is kept in a separate module is to easily
-          be able to provide a stub module that doesn't alter system
-          state at all (for unit tests)
-    """
-    return os.unlink(path)
-
-
-def path_exists(path):
-    """Returns if path exists
-
-    Note: The reason this is kept in a separate module is to easily
-          be able to provide a stub module that doesn't alter system
-          state at all (for unit tests)
-    """
-    return os.path.exists(path)
-
-
-def find_disk(virt_dom):
+def find_disk(guest):
     """Find root device path for instance
 
     May be file or device
     """
-    xml_desc = virt_dom.XMLDesc(0)
-    domain = etree.fromstring(xml_desc)
-    os_type = domain.find('os/type').text
-    driver = None
-    if CONF.libvirt.virt_type == 'lxc':
-        filesystem = domain.find('devices/filesystem')
-        driver = filesystem.find('driver')
+    guest_config = guest.get_config()
 
-        source = filesystem.find('source')
-        disk_path = source.get('dir')
+    disk_format = None
+    if guest_config.virt_type == 'lxc':
+        filesystem = next(d for d in guest_config.devices
+                          if isinstance(d, vconfig.LibvirtConfigGuestFilesys))
+        disk_path = filesystem.source_dir
         disk_path = disk_path[0:disk_path.rfind('rootfs')]
         disk_path = os.path.join(disk_path, 'disk')
-    elif CONF.libvirt.virt_type == 'parallels' and os_type == vm_mode.EXE:
-        filesystem = domain.find('devices/filesystem')
-        driver = filesystem.find('driver')
-
-        source = filesystem.find('source')
-        disk_path = source.get('file')
+    elif (guest_config.virt_type == 'parallels' and
+          guest_config.os_type == obj_fields.VMMode.EXE):
+        filesystem = next(d for d in guest_config.devices
+                          if isinstance(d, vconfig.LibvirtConfigGuestFilesys))
+        disk_format = filesystem.driver_type
+        disk_path = filesystem.source_file
     else:
-        disk = domain.find('devices/disk')
-        driver = disk.find('driver')
-
-        source = disk.find('source')
-        disk_path = source.get('file') or source.get('dev')
-        if not disk_path and CONF.libvirt.images_type == 'rbd':
-            disk_path = source.get('name')
+        disk = next(d for d in guest_config.devices
+                    if isinstance(d, vconfig.LibvirtConfigGuestDisk))
+        disk_format = disk.driver_format
+        disk_path = disk.source_path if disk.source_type != 'mount' else None
+        if not disk_path and disk.source_protocol == 'rbd':
+            disk_path = disk.source_name
             if disk_path:
                 disk_path = 'rbd:' + disk_path
 
@@ -371,15 +368,11 @@ def find_disk(virt_dom):
         raise RuntimeError(_("Can't retrieve root device path "
                              "from instance libvirt configuration"))
 
-    if driver is not None:
-        format = driver.get('type')
-        # This is a legacy quirk of libvirt/xen. Everything else should
-        # report the on-disk format in type.
-        if format == 'aio':
-            format = 'raw'
-    else:
-        format = None
-    return (disk_path, format)
+    # This is a legacy quirk of libvirt/xen. Everything else should
+    # report the on-disk format in type.
+    if disk_format == 'aio':
+        disk_format = 'raw'
+    return (disk_path, disk_format)
 
 
 def get_disk_type_from_path(path):
@@ -415,18 +408,29 @@ def get_fs_info(path):
             'used': used}
 
 
-def fetch_image(context, target, image_id):
-    """Grab image."""
-    images.fetch_to_raw(context, image_id, target)
+def fetch_image(context, target, image_id, trusted_certs=None):
+    """Grab image.
+
+    :param context: nova.context.RequestContext auth request context
+    :param target: target path to put the image
+    :param image_id: id of the image to fetch
+    :param trusted_certs: optional objects.TrustedCerts for image validation
+    """
+    images.fetch_to_raw(context, image_id, target, trusted_certs)
 
 
-def fetch_raw_image(context, target, image_id):
+def fetch_raw_image(context, target, image_id, trusted_certs=None):
     """Grab initrd or kernel image.
 
     This function does not attempt raw conversion, as these images will
     already be in raw format.
+
+    :param context: nova.context.RequestContext auth request context
+    :param target: target path to put the image
+    :param image_id: id of the image to fetch
+    :param trusted_certs: optional objects.TrustedCerts for image validation
     """
-    images.fetch(context, image_id, target)
+    images.fetch(context, image_id, target, trusted_certs)
 
 
 def get_instance_path(instance, relative=False):
@@ -488,26 +492,116 @@ def get_arch(image_meta):
         if image_arch is not None:
             return image_arch
 
-    return arch.from_host()
+    return obj_fields.Architecture.from_host()
 
 
 def is_mounted(mount_path, source=None):
     """Check if the given source is mounted at given destination point."""
-    try:
-        check_cmd = ['findmnt', '--target', mount_path]
-        if source:
-            check_cmd.extend(['--source', source])
+    if not os.path.ismount(mount_path):
+        return False
 
-        utils.execute(*check_cmd)
+    if source is None:
         return True
-    except processutils.ProcessExecutionError:
-        return False
-    except OSError as exc:
-        # info since it's not required to have this tool.
-        if exc.errno == errno.ENOENT:
-            LOG.info(_LI("findmnt tool is not installed"))
-        return False
+
+    with open('/proc/mounts', 'r') as proc_mounts:
+        mounts = [mount.split() for mount in proc_mounts.readlines()]
+        return any(mnt[0] == source and mnt[1] == mount_path for mnt in mounts)
 
 
 def is_valid_hostname(hostname):
     return re.match(r"^[\w\-\.:]+$", hostname)
+
+
+def version_to_string(version):
+    """Returns string version based on tuple"""
+    return '.'.join([str(x) for x in version])
+
+
+def cpu_features_to_traits(features):
+    """Returns this driver's CPU traits dict where keys are trait names from
+    CPU_TRAITS_MAPPING, values are boolean indicates whether the trait should
+    be set in the provider tree.
+    """
+    traits = {trait_name: False for trait_name in CPU_TRAITS_MAPPING.values()}
+    for f in features:
+        if f in CPU_TRAITS_MAPPING:
+            traits[CPU_TRAITS_MAPPING[f]] = True
+
+    return traits
+
+
+def get_cpu_model_from_arch(arch):
+    mode = 'qemu64'
+    if arch == obj_fields.Architecture.I686:
+        mode = 'qemu32'
+    elif arch == obj_fields.Architecture.PPC64LE:
+        mode = 'POWER8'
+    return mode
+
+
+def get_machine_type(image_meta):
+    """The guest machine type can be set as an image metadata property, or
+    otherwise based on architecture-specific defaults. If no defaults are
+    found then None will be returned. This will ultimately lead to QEMU using
+    its own default which is currently the 'pc' machine type.
+    """
+    if image_meta.properties.get('hw_machine_type') is not None:
+        return image_meta.properties.hw_machine_type
+
+    # If set in the config, use that as the default.
+    return get_default_machine_type(get_arch(image_meta))
+
+
+def get_default_machine_type(arch):
+    # NOTE(lyarwood): Values defined in [libvirt]/hw_machine_type take
+    # precedence here if available for the provided arch.
+    for mapping in CONF.libvirt.hw_machine_type or {}:
+        host_arch, _, machine_type = mapping.partition('=')
+        if machine_type == '':
+            LOG.warning("Invalid hw_machine_type config value %s", mapping)
+        elif host_arch == arch:
+            return machine_type
+    # NOTE(kchamart): For ARMv7 and AArch64, use the 'virt' board as the
+    # default machine type.  It is the recommended board, which is designed
+    # to be used with virtual machines.  The 'virt' board is more flexible,
+    # supports PCI, 'virtio', has decent RAM limits, etc.
+    #
+    # NOTE(sean-k-mooney): Nova's default for x86 is still 'pc', so
+    # use that, not 'q35', for x86_64 and i686.
+    #
+    # NOTE(aspiers): If you change this, don't forget to update the
+    # docs and metadata for hw_machine_type in glance.
+    default_mtypes = {
+        obj_fields.Architecture.ARMV7: "virt",
+        obj_fields.Architecture.AARCH64: "virt",
+        obj_fields.Architecture.S390: "s390-ccw-virtio",
+        obj_fields.Architecture.S390X: "s390-ccw-virtio",
+        obj_fields.Architecture.I686: "pc",
+        obj_fields.Architecture.X86_64: "pc",
+    }
+    return default_mtypes.get(arch)
+
+
+def mdev_name2uuid(mdev_name):
+    """Convert an mdev name (of the form mdev_<uuid_with_underscores>) to a
+    uuid (of the form 8-4-4-4-12).
+    """
+    return str(uuid.UUID(mdev_name[5:].replace('_', '-')))
+
+
+def mdev_uuid2name(mdev_uuid):
+    """Convert an mdev uuid (of the form 8-4-4-4-12) to a name (of the form
+    mdev_<uuid_with_underscores>).
+    """
+    return "mdev_" + mdev_uuid.replace('-', '_')
+
+
+def get_flags_by_flavor_specs(flavor):
+    req_spec = objects.RequestSpec(flavor=flavor)
+    resource_request = scheduler_utils.ResourceRequest(req_spec)
+    required_traits = resource_request.all_required_traits
+
+    flags = [TRAITS_CPU_MAPPING[trait] for trait in required_traits
+             if trait in TRAITS_CPU_MAPPING]
+
+    return set(flags)
